(** A graph-tableau-based decision procedure framework for coalgebraic logics.
    @author Florian Widmann
 *)


open CoAlgMisc

module M = Minisat

module CU = CoolUtils

type sortTable = CoAlgMisc.sortTable

type nomTable = string -> CoAlgFormula.sort option

type assumptions = CoAlgFormula.sortedFormula list

type input = sortTable * nomTable * assumptions * CoAlgFormula.sortedFormula

exception ReasonerError of string

(*****************************************************************************)
(*                     Propagation of Satisfiability                         *)
(*****************************************************************************)

let propSatFindSucc setCnstr cset =
  if csetHasDot cset then false
  else
    match graphFindCnstr cset with
    | None -> raise (ReasonerError "?")
    | Some SatC -> true
    | Some (OpenC _) -> setMemCnstr setCnstr cset
    | Some (UnexpandedC _)
    | Some UnsatC -> false


let rec propSat setStates setCores setCnstr = function
  | [] -> ()
  | propElem::tl ->
      let tl1 =
        match propElem with
        | UState (state, _) ->
            if setMemState setStates state then
              let cnstrs = stateGetConstraints state in
              let isStillOpen = cssExists (fun c -> propSatFindSucc setCnstr c) cnstrs in
              if isStillOpen then () else setRemoveState setStates state
            else ();
            tl
        | UCore (core, _) ->
            if setMemCore setCores core then
              let cnstrs = coreGetConstraints core in
              let isStillOpen = cssExists (fun c -> propSatFindSucc setCnstr c) cnstrs in
              if isStillOpen then tl else begin
                setRemoveCore setCores core;
                let cnstrPar = coreGetConstraintParents core in
                List.fold_left (fun acc cset -> (UCnstr cset)::acc) tl cnstrPar
              end
            else tl
        | UCnstr cset ->
            if setMemCnstr setCnstr cset then begin
              setRemoveCnstr setCnstr cset;
              match graphFindCnstr cset with
              | Some (OpenC nodes) ->
                  let prop acc node =
                    match node with
                    | Core core -> (UCore (core, true))::acc
                    | State state -> (UState (state, None))::acc
                  in
                  List.fold_left prop tl nodes
              | _ -> assert false
            end else tl
      in
      propSat setStates setCores setCnstr tl1

let propagateSat () =
  let setStates = setEmptyState () in
  let setCores = setEmptyCore () in
  let setCnstr = setEmptyCnstr () in
  let fktS1 state =
    match stateGetStatus state with
    | Expandable
    | Open -> setAddState setStates state
    | Unsat
    | Sat -> ()
  in
  graphIterStates fktS1;
  let fktC1 core =
    match coreGetStatus core with
    | Expandable
    | Open -> setAddCore setCores core
    | Unsat
    | Sat -> ()
  in
  graphIterCores fktC1;
  let cfkt1 cset cnstr =
    if csetHasDot cset then ()
    else
      match cnstr with
      | OpenC _ -> setAddCnstr setCnstr cset
      | UnsatC
      | SatC
      | UnexpandedC _ -> ()
  in
  graphIterCnstrs cfkt1;
  graphIterStates (fun state -> propSat setStates setCores setCnstr [UState (state, None)]);
  graphIterCores (fun core -> propSat setStates setCores setCnstr [UCore (core, false)]);
  setIterState (fun state -> stateSetStatus state Sat) setStates;
  let setCoreSat core =
    coreSetStatus core Sat;
    if core == graphGetRoot () then raise (CoAlg_finished true) else ()
  in
  setIterCore setCoreSat setCores;
  setIterCnstr (fun cset -> graphReplaceCnstr cset SatC) setCnstr


let propagateSatMu () =
  let setFinishingStates = setEmptyState () in
  let setFinishingCores = setEmptyCore () in
  let setStates = setEmptyState () in
  let setCores = setEmptyCore () in
  let emptySet = bsetMakeRealEmpty () in
  let openstates = ref 0 in

  (* Collect two sets of nodes. All nodes that may be satisfiable
   * after this iteration are collected into setStates/setCores.
   *
   * As every cycle containing a node with empty focus or an
   * satisfiable node should be considered satisfiable, collect these
   * decisive nodes into setFinishingStates/setFinishingCores
   *
   * This also marks in trivial cases nodes as satisfiable.
   *)
  let stateCollector state =
    match stateGetStatus state with
    | Unsat -> ()
    | Sat ->
       setAddState setStates state;
      setAddState setFinishingStates state
    | Expandable -> ()
    | Open ->
       openstates := !openstates + 1;
       if List.length (stateGetRules state) == 0 || (* States with no rules are satisfiable *)
         bsetCompare (bsetMake ()) (stateGetBs state) == 0 (* KD generates nodes with just True as formula *)
       then begin
         setAddState setFinishingStates state;
         stateSetStatus state Sat
       end else begin
         setAddState setStates state;
         if bsetCompare (stateGetDeferral state) emptySet == 0
         then begin
           setAddState setFinishingStates state
         end
         else ()
       end

  (* As it is enough for a core to have one successfull child, we can
   * also handle (some) expandable cores.
   *)
  and coreCollector core =
    match coreGetStatus core with
    | Unsat -> ()
    | Sat ->
       setAddCore setCores core;
       setAddCore setFinishingCores core
    | Expandable
    | Open ->
       setAddCore setCores core;
      if bsetCompare (coreGetDeferral core) emptySet == 0
      then begin
        setAddCore setFinishingCores core
      end
      else ()
  in
  graphIterStates stateCollector;
  graphIterCores coreCollector;

  setPropagationCounter !openstates;

  (* In a fixpoint the set called setStates / setCores is narrowed
   * down.
   *
   * In each step only those states and cores are retained in setStates
   * / setCores which reach one of setFinishing{States,Cores} in
   * finitely many steps. This new set of States / Cores is collected
   * as allowed{States,Cores} during each fixpoint iteration.
   *
   * Only those finishing nodes are retained that have allowed or
   * Successfull Children.
   *)
  let rec fixpointstep setStates setCores =
    let allowedStates = setEmptyState () in
    let allowedCores = setEmptyCore () in

    let rec visitParentStates (core : core) : unit =
      if not (setMemCore allowedCores core)
      then begin
        setAddCore allowedCores core;
        let verifyParent (state,_) =
          let rules = stateGetRules state in
          let ruleiter (dependencies, corelist) =
            List.exists (fun (core : core) -> setMemCore allowedCores core ||
                                                coreGetStatus core == Sat)
              corelist
          in
          if List.for_all ruleiter rules
          then visitParentCores state
        in
        List.iter verifyParent (coreGetParents core)
      end

    and visitParentCores (state : state) : unit =
      if not (setMemState allowedStates state)
      then begin
        setAddState allowedStates state;
        let verifyParent core =
          let acceptable =
            List.exists (fun (state : state) -> setMemState allowedStates state ||
                                                  stateGetStatus state == Sat)
              (coreGetChildren core)
          in
          if acceptable
          then visitParentStates core
        in
        List.iter verifyParent (stateGetParents state)
      end
    in

    (* All rule applications need to still be potentially Sat for a
     * finishing State to be a valid startingpoint for this fixpoint.
     *)
    let checkFinishingState (state : state) =
      let ruleiter (dependencies, corelist) : bool =
        List.for_all (fun (core : core) -> coreGetStatus core == Unsat ||
                                             coreGetStatus core == Expandable ||
                                               not (setMemCore setCores core)) corelist
      in
      if not (List.exists ruleiter (stateGetRules state)) then begin
        visitParentCores state
      end

    (* There needs to be a State still potentially Sat for this core
     * to be considered for the fixpoint
     *)
    and checkFinishingCore (core : core) =
      if not (List.for_all (fun (state : state) -> stateGetStatus state == Unsat ||
                                                     stateGetStatus state == Expandable ||
                                                       not (setMemState setStates state))
                           (coreGetChildren core))
      then begin
        visitParentStates core
      end
    in
    setIterState checkFinishingState setFinishingStates;
    setIterCore checkFinishingCore setFinishingCores;


    if (setLengthState setStates) = (setLengthState allowedStates) &&
         (setLengthCore setCores) = (setLengthCore allowedCores)
    then begin
        setIterState (fun state -> stateSetStatus state Sat) setStates;
        setIterCore (fun core -> coreSetStatus core Sat; if core == graphGetRoot ()
                                                         then raise (CoAlg_finished true)
                                                         else ()) setCores;
      end else
      fixpointstep allowedStates allowedCores
  in
  fixpointstep setStates setCores


(*****************************************************************************)
(*                     Propagation of Unsatisfiability                       *)
(*****************************************************************************)

let isConstraintUnsat cset =
  match graphFindCnstr cset with
  | None -> assert false
  | Some UnsatC -> true
  | _ -> false

let fhtMustFind fht f =
  match fhtFind fht f with
  | Some l -> l
  | None -> assert false

let lhtMustFind lht l =
  match lhtFind lht l with
  | Some f -> f
  | None -> assert false


(* Gets a list of Things we know are unsatisfiable and propagates this
   information backwards *)
let rec propagateUnsat = function
  | [] -> ()
  | propElem::tl ->
      let tl1 =
        match propElem with
        | UState (state, idxopt) -> begin
            match stateGetStatus state with
            | Unsat -> tl
            | Sat -> raise (ReasonerError ("Propagation tells state "
                                           ^(string_of_int (stateGetIdx state))
                                           ^" would be unsat, but it is already sat"))
            | Open
            | Expandable ->
                let mbs =
                  match idxopt with
                  | None -> stateGetBs state
                  | Some idx ->
                      let (dep, children) = stateGetRule state idx in
                      let getBs core =
                        assert (coreGetStatus core = Unsat);
                        coreGetBs core
                      in
                      dep (List.map getBs children)
                in
                stateSetBs state mbs;
                stateSetStatus state Unsat;
                let prop acc core =
                  let turnsUnsat =
                    match coreGetStatus core with
                    | Open -> List.for_all (fun s -> stateGetStatus s = Unsat)
                                           (coreGetChildren core)
                    | Expandable
                    | Unsat
                    | Sat -> false
                  in
                  if turnsUnsat then (UCore (core, false))::acc else acc
                in
                List.fold_left prop tl (stateGetParents state)
          end
        | UCore (core, comesFromCnstr) -> begin
            match coreGetStatus core with
            | Unsat -> tl
            | Sat -> raise (ReasonerError ("Propagation tells core "
                                           ^(string_of_int (coreGetIdx core))
                                           ^" would be unsat, but it is already sat"))
            | Open
            | Expandable ->
                let mbs =
                  if comesFromCnstr then coreGetBs core
                  else
                    let bs = coreGetBs core in
                    let Some solver = coreGetSolver core in
                    let fht = coreGetFht core in
                    let lht = lhtInit () in
                    let addLits f acc =
                      let lit = fhtMustFind fht f in
                      lhtAdd lht lit f;
                      lit::acc
                    in
                    let litset = bsetFold addLits bs [] in
                    let addClauses state =
                      let cbs = stateGetBs state in
                      let clause = bsetFold (fun f acc -> (M.neg_lit (fhtMustFind fht f))::acc) cbs [] in
                      let okay = M.add_clause solver clause in
                      assert okay
                    in
                    List.iter addClauses (coreGetChildren core);
                    let isSat = M.invoke_solver solver litset in
                    assert (not isSat);
                    let res = bsetMake () in
                    let confls = M.get_conflict solver in
                    List.iter (fun l -> bsetAdd res (lhtMustFind lht l)) confls;
                    res
                in
                coreSetBs core mbs;
                coreSetStatus core Unsat;
                coreDeallocateSolver core;
                if core == graphGetRoot () then raise (CoAlg_finished false) else ();
                let prop acc (state, idx) =
                  let turnsUnsat =
                    match stateGetStatus state with
                    | Open
                    | Expandable ->
                        List.for_all (fun c -> coreGetStatus c = Unsat) (snd (stateGetRule state idx))
                    | Unsat
                    | Sat -> false
                  in
                  if turnsUnsat then (UState (state, Some idx))::acc else acc
                in
                let tl2 = List.fold_left prop tl (coreGetParents core) in
                List.fold_left (fun acc cset -> (UCnstr cset)::acc) tl2 (coreGetConstraintParents core)
          end
        | UCnstr cset ->
            match graphFindCnstr cset with
            | None
            | Some SatC -> assert false
            | Some UnsatC -> tl
            | Some (UnexpandedC nodes)
            | Some (OpenC nodes) ->
                graphReplaceCnstr cset UnsatC;
                let prop acc node =
                  match node with
                  | State state ->
                      let turnsUnsat =
                        match stateGetStatus state with
                        | Expandable
                        | Open -> cssForall isConstraintUnsat (stateGetConstraints state)
                        | Unsat
                        | Sat -> false
                      in
                      if turnsUnsat then (UState (state, None))::acc else acc
                  | Core core ->
                      let turnsUnsat =
                        match coreGetStatus core with
                        | Expandable
                        | Open -> cssForall isConstraintUnsat (coreGetConstraints core)
                        | Unsat
                        | Sat -> false
                      in
                      if turnsUnsat then (UCore (core, true))::acc else acc
                in
                List.fold_left prop tl nodes
      in
      propagateUnsat tl1

let propagateUnsatMu () =
  let setFinishingStates = setEmptyState () in
  let setFinishingCores = setEmptyCore () in
  let setStates = setEmptyState () in
  let setCores = setEmptyCore () in
  let emptySet = bsetMakeRealEmpty () in

  (* Collect two sets of nodes. All nodes that may be unsatisfiable
   * after this iteration are collected into setStates/setCores.
   *
   * Nodes reaching Points with empty focus as well as Expandable nodes
   * (if not Unsat) can not be declared Unsat so we collect these into
   * setFinishingStates/setFinishingCores.
   *)
  let stateCollector state =
    match stateGetStatus state with
    | Unsat -> ()
    | Sat
    | Expandable ->
       setAddState setStates state;
      setAddState setFinishingStates state
    | Open ->
       setAddState setStates state;
      if [] = (stateGetRules state)
      then begin
        stateSetStatus state Sat;
        setAddState setFinishingStates state
      end
      else ();
      if bsetCompare (stateGetDeferral state) emptySet == 0
      then begin
        setAddState setFinishingStates state
      end
      else ()

  and coreCollector core =
    match coreGetStatus core with
    | Unsat -> ()
    | Expandable
    | Sat ->
       setAddCore setCores core;
      setAddCore setFinishingCores core
    | Open ->
       setAddCore setCores core;
      if bsetCompare (coreGetDeferral core) emptySet == 0
      then begin
        setAddCore setFinishingCores core
      end
      else ()
  in
  graphIterStates stateCollector;
  graphIterCores coreCollector;

  (* In a fixpoint the set called setFinishingStates/setFinishingCores
   * is narrowed down
   *
   * In each iteration we start with all Nodes and remove all that can
   * reach a finishing Node. We then remove all finishing Nodes from the
   * respective set which are becoming Unsat and start with the next
   * iteration until the fixpoint is reached
   *
   *
   *)
  let rec fixpointstep setPrevUnsatStates setPrevUnsatCores =
    let setUnsatStates = setCopyState setStates in
    let setUnsatCores = setCopyCore setCores in

    let rec visitParentStates (core : core) : unit =
      if setMemCore setUnsatCores core
      then begin
        setRemoveCore setUnsatCores core;
        let verifyParent (state,_) =
          let rules = stateGetRules state in
          let ruleiter (dependencies, corelist) =
            List.exists (fun (core : core) -> not (setMemCore setUnsatCores core) ||
                                                coreGetStatus core == Sat)
              corelist
          in
          if List.for_all ruleiter rules
          then visitParentCores state
        in
        List.iter verifyParent (coreGetParents core)
      end

    and visitParentCores (state : state) : unit =
      if setMemState setUnsatStates state
      then begin
        setRemoveState setUnsatStates state;
        let verifyParent core =
          let acceptable =
            List.exists (fun (state : state) -> not (setMemState setUnsatStates state) ||
                                                  stateGetStatus state == Sat)
              (coreGetChildren core)
          in
          if acceptable
          then visitParentStates core
        in
        List.iter verifyParent (stateGetParents state)
      end
    in

    (* All rule applications need to still be potentially Sat for a
     * finishing State to be a valid startingpoint for this fixpoint.
     *)
    let checkFinishingState (state : state) =
      let ruleiter (dependencies, corelist) : bool =
        List.for_all (fun (core : core) -> coreGetStatus core == Unsat ||
                                             setMemCore setPrevUnsatCores core) corelist
      in
      if not (List.exists ruleiter (stateGetRules state)) ||
           stateGetStatus state == Expandable then begin
        visitParentCores state
      end

    (* There needs to be a State still potentially Sat for this core
     * to be considered for the fixpoint
     *)
    and checkFinishingCore (core : core) =
      if not (List.for_all (fun (state : state) -> stateGetStatus state == Unsat ||
                                                     setMemState setPrevUnsatStates state)
                (coreGetChildren core)) ||
        coreGetStatus core == Expandable
      then begin
        visitParentStates core
      end
    in
    setIterState checkFinishingState setFinishingStates;
    setIterCore checkFinishingCore setFinishingCores;

    if (setLengthState setPrevUnsatStates) = (setLengthState setUnsatStates) &&
      (setLengthCore setPrevUnsatCores) = (setLengthCore setUnsatCores)
    then begin
        setIterState (fun state -> stateSetStatus state Unsat) setUnsatStates;
        setIterCore (fun core -> propagateUnsat [UCore (core, false)]) setUnsatCores;
      end else
      fixpointstep setUnsatStates setUnsatCores
  in
  fixpointstep (setEmptyState ()) (setEmptyCore ())


(*****************************************************************************)
(*                   Propagation of Nominal Constraints                      *)
(*****************************************************************************)

let extractAts sort bs =
  let ats = csetMake () in
  let procNom nom f =
    match lfGetType sort f with
    | AndF
    | OrF -> ()
    | TrueFalseF
    | AtF -> ()
    | _ -> csetAdd ats (lfGetAt (sort, nom) f)
  in
  let getAts f =
    match lfGetType sort f with
    | AtF -> csetAdd ats (lfToAt sort f)
    | NomF -> bsetIter (fun f1 -> procNom f f1) bs
    | _ -> ()
  in
  bsetIter getAts bs;
  ats

let detConstraintsState state =
  let procRule accR (_, chldrn) =
    let procChldrn accC core =
      if coreGetStatus core = Unsat then accC
      else cssUnion accC (coreGetConstraints core)
    in
    let orset = List.fold_left procChldrn cssEmpty chldrn in
    let procOrset csetO accO =
      let mkUnion csetU accU =
        let cset = csetUnion csetO csetU in
        cssAdd cset accU
      in
      cssFold mkUnion accR accO
    in
    cssFold procOrset orset cssEmpty
  in
  let sort = stateGetSort state in
  let bs = stateGetBs state in
  let ats = extractAts sort bs in
  if stateGetStatus state = Expandable then csetAddDot ats else ();
  List.fold_left procRule (cssSingleton ats) (stateGetRules state)

let detConstraintsCore core =
  let sort = coreGetSort core in
  let bs = coreGetBs core in
  let ats = extractAts sort bs in
  let addCnstr acc state =
    if stateGetStatus state = Unsat then acc
    else
      let csets = stateGetConstraints state in
      (* cssFold (fun c a -> cssAdd (csetUnion c ats) a) csets acc *)
      cssUnion csets acc
  in
  let initInner =
    if coreGetStatus core = Expandable then
      (* let cset = csetCopy ats in *)
      let cset = ats in
      csetAddDot cset;
      cssSingleton cset
    else cssEmpty
  in
  List.fold_left addCnstr initInner (coreGetChildren core)

let rec propNom = function
  | [] -> ()
  | node::tl ->
      let tl1 =
        match node with
        | State state ->
            if stateGetStatus state = Unsat then tl
            else
              let csets = detConstraintsState state in
              let oldcsets = stateGetConstraints state in
              if cssEqual csets oldcsets then tl
              else begin
                stateSetConstraints state csets;
                List.fold_left (fun acc c -> (Core c)::acc) tl (stateGetParents state)
              end
        | Core core ->
            if coreGetStatus core = Unsat then tl
            else
              let csets = detConstraintsCore core in
              let oldcsets = coreGetConstraints core in
              if cssEqual csets oldcsets then tl
              else begin
                coreSetConstraints core csets;
                List.fold_left (fun acc (s, _) -> (State s)::acc) tl (coreGetParents core)
              end
      in
      propNom tl1

let updateConstraints node elemU csets =
  let update cset acc =
    let cnstr =
      match graphFindCnstr cset with
      | Some cn -> cn
      | None ->
          let cn = UnexpandedC [] in
          graphInsertCnstr cset cn;
          queueInsertCnstr cset;
          cn
    in
    match cnstr with
    | UnsatC -> acc
    | UnexpandedC parents ->
        graphReplaceCnstr cset (UnexpandedC (node::parents));
        false
    | OpenC parents ->
        graphReplaceCnstr cset (OpenC (node::parents));
        false
    | SatC -> false
  in
  let isUnsat = cssFold update csets true in
  if isUnsat then propagateUnsat [elemU] else ()

let propagateNominals () =
  let init = cssSingleton (csetMake ()) in
  graphIterStates (fun s -> if stateGetStatus s = Unsat then () else stateSetConstraints s init);
  graphIterCores (fun c -> if coreGetStatus c = Unsat then () else coreSetConstraints c init);
  graphIterStates (fun s -> propNom [State s]);
  graphIterCores (fun c -> propNom [Core c]);
  graphClearCnstr ();
  let fktS state =
    if stateGetStatus state = Unsat then ()
    else updateConstraints (State state) (UState (state, None)) (stateGetConstraints state)
  in
  graphIterStates fktS;
  let fktC core =
    if coreGetStatus core = Unsat then ()
    else updateConstraints (Core core) (UCore (core, true)) (coreGetConstraints core)
  in
  graphIterCores fktC


(*****************************************************************************)
(*                           Node Expansions                                 *)
(*****************************************************************************)

let getLit sort fht solver f =
  match fhtFind fht f with
  | Some lit -> lit
  | None ->
      let var = M.new_variable solver in
      let lit = M.var_to_lit var true in
      fhtAdd fht f lit;
      let () =
        match lfGetNeg sort f with
        | None -> ()
        | Some nf ->
            let nlit = M.neg_lit lit in
            fhtAdd fht nf nlit;
      in
      lit

let newCore sort bs defer =
  (* when creating a now core from a set of formulas bs
        bs = { x_1, ...., x_n }
           = "x_1 /\ .... /\ x_n"
     Then we do this by:

        1. creating a new literal in the sat solver for each
           "boolean subformula" from the x_i. "boolean subformula" means that
           this subformula is not hidden under modalities but only under
           boolean connectives (/\, \/).
        2. encoding the relation between a formula and its intermediate boolean
           subformulas by a clause:
  *)
  let fht = fhtInit () in
  let solver = M.new_solver () in
  let rec addClauses f =
    (* step 1: create a literal in the satsolver representing the formula f *)
    let lf = getLit sort fht solver f in
    (* step 2: encode relation to possible subformulas *)
    match lfGetType sort f with
    | OrF ->
        (*
           case 2.(a)   f =  f1 \/ f2
                   fill fht such that it says:

                        f  |---> lf
                        f1 |---> lf1
                        f2 |---> lf2
        *)
        let f1 = lfGetDest1 sort f in
        let f2 = lfGetDest2 sort f in
        addClauses f1;
        addClauses f2;
        let lf1 = fhtMustFind fht f1 in
        let lf2 = fhtMustFind fht f2 in
        (*
            encode the structure of f by adding the clause (lf -> lf1 \/ lf2)
        *)
        let okay = M.add_clause solver [M.neg_lit lf; lf1; lf2] in
        assert (okay)
    | AndF ->
        (*
           case 2.(b)   f =  f1 /\ f2
                   fill fht such that it says:

                        f  |---> lf
                        f1 |---> lf1
                        f2 |---> lf2
        *)
        let f1 = lfGetDest1 sort f in
        let f2 = lfGetDest2 sort f in
        addClauses f1;
        addClauses f2;
        let lf1 = fhtMustFind fht f1 in
        let lf2 = fhtMustFind fht f2 in
        (*
            encode the structure of f by adding the clauses
                    (lf -> lf1) /\ (lf -> lf2)
        *)
        let okay1 = M.add_clause solver [M.neg_lit lf; lf1] in
        assert (okay1);
        let okay2 = M.add_clause solver [M.neg_lit lf; lf2] in
        assert (okay2)
    | MuF
    | NuF ->
       (*
          Dest of a fixpoint is it's unfolded version. This adds just
          an simple forward implication that could be optimised out
          though not without nontrivial transformation of code

            f = μ X . φ   |---> lf
            φ[X |-> f]    |---> lf1

          Then adding lf -> lf1 to minisat
        *)
       let f1 = lfGetDest1 sort f in
       addClauses f1;
       let lf1 = fhtMustFind fht f1 in
       let okay1 = M.add_clause solver [M.neg_lit lf; lf1] in
       assert (okay1);
    | _ -> ()
        (* case 2.(c)
            We don't need to do anything except adding f to the fht
        *)
  in
  bsetIter addClauses bs;
  coreMake sort bs defer (Some solver) fht

let getNextState core =
  (* Create a new state, which is obtained by a satisfying assignment of the
     literals of the minisat solver.
     In addition to that, feed the negation of the satisfying assignment back
     to the minisat solver in order to obtain different solutions on successive
     minisat calls.
  *)
  let bs = coreGetBs core in
  let refocusing = bsetCompare (bsetMakeRealEmpty ()) (coreGetDeferral core) = 0 in
  let deferralS =
    if refocusing then
      bsetCopy bs
    else
      coreGetDeferral core
  in
  let fht = coreGetFht core in
  let litset = bsetFold (fun f acc -> (fhtMustFind fht f)::acc) bs [] in
  let Some solver = coreGetSolver core in
  let isSat = M.invoke_solver solver litset in
  (* Clearly, if the current core is unsatisfiable, no further child state can
     be created *)
  if not isSat then None
  else
    let sort = coreGetSort core in
    (* create a new set of formulas newbs with the property:
       if the conjunction of newbs is satisfiable, then the present core is
       satisfiable.
    *)
    let newbs = bsetMake () in
    let newdefer = bsetMakeRealEmpty () in
    (* if newbs = { l_1, .... l_n }, i.e.

            newbs = l_1 /\ ... /\ l_n

       then we can get essentially different satisfying assignments of bs by
       adding another clause

            clause = ¬ newbs = ¬l_1 \/ ... \/ ¬l_n

       By mkExclClause, newbs is filled, and clause is built in the accumulator acc.
    *)
    let rec mkExclClause deferral f acc =
      (* f is a formula that is true in the current satisfying assignment *)
      match lfGetType sort f with
      | OrF ->
          (* if f is true and a disjunction, then one of its disjuncts is true *)
          let f1 = lfGetDest1 sort f in
          let lf1 = fhtMustFind fht f1 in
          (* if the first disjunct f1 is true, then we need to add subformulas
             of f1 to newbs&clause *)
          if M.literal_status solver lf1 = M.LTRUE then mkExclClause deferral f1 acc
          else
            (* otherwise f2 must be true *)
            let f2 = lfGetDest2 sort f in
            let lf2 = fhtMustFind fht f2 in
            assert (M.literal_status solver lf2 = M.LTRUE);
            mkExclClause deferral f2 acc
      | AndF ->
          (* if the true f is a conjuction, then both conjunctions must be true *)
          let f1 = lfGetDest1 sort f in
          let lf1 = fhtMustFind fht f1 in
          assert (M.literal_status solver lf1 = M.LTRUE);
          let acc1 = mkExclClause deferral f1 acc in
          let f2 = lfGetDest2 sort f in
          let lf2 = fhtMustFind fht f2 in
          assert (M.literal_status solver lf2 = M.LTRUE);
          mkExclClause deferral f2 acc1
      | MuF
      | NuF ->
         let f1 = lfGetDest1 sort f in
         mkExclClause deferral f1 acc
      | _ ->
         (* if f is a trivial formula or modality, then add it ... *)
         (* ... to newbs *)
         bsetAdd newbs f;
         let defercandidate = lfGetDeferral sort f in
         (if (defercandidate != (Hashtbl.hash "ε") &&
                (refocusing || deferral = defercandidate)) then
            bsetAdd newdefer f);
         (* ... and to the new clause *)
         (M.neg_lit (fhtMustFind fht f))::acc
    in
    let init_clause f acc =
      let deferral =
        (* for each deferral determine which variable it belongs to *)
        if bsetMem deferralS f then (
          lfGetDeferral sort f)
        else
          (Hashtbl.hash "ε")
      in
      mkExclClause deferral f acc
    in
    let clause = bsetFold init_clause bs [] in
    let okay = M.add_clause solver clause in
    assert (okay);
    Some (sort, newbs, newdefer)

let newState sort bs defer =
  let (func, sl) = !sortTable.(sort) in
  let producer = CoAlgLogics.getExpandingFunctionProducer func in
  let exp = producer sort bs defer sl in
  stateMake sort bs defer exp

let insertState parent sort bs defer =
  let child =
    match graphFindState sort (bs, defer) with
    | None ->
        let s = newState sort bs defer in
        graphInsertState sort (bs, defer) s;
        queueInsertState s;
        s
    | Some s -> s
  in
  coreAddChild parent child;
  stateAddParent child parent

let expandCore core =
  match getNextState core with
  | Some (sort, bs, defer) ->
      insertState core sort bs defer;
      queueInsertCore core
  | None ->
      let isUnsat = List.for_all (fun s -> stateGetStatus s = Unsat) (coreGetChildren core) in
      if isUnsat
      then
          propagateUnsat [UCore (core, false)]
      else
          coreSetStatus core Open


let insertCore sort bs defer =
  match graphFindCore sort (bs, defer) with
  | None ->
     let c = newCore sort bs defer in
     graphInsertCore sort (bs, defer) c;
     queueInsertCore c;
     c
  | Some c -> c

let insertRule state dep chldrn =
  let chldrn = listFromLazylist chldrn in
  let insert (isUns, acc) (sort, bs, defer) =
    let bs1 = bsetAddTBox sort bs in
    let core = insertCore sort bs1 defer in
    let isUns1 = if coreGetStatus core = Unsat then isUns else false in
    (isUns1, core::acc)
  in
  let (isUnsat, children) = List.fold_left insert (true, []) chldrn in
  let idx = stateAddRule state dep (List.rev children) in
  List.iter (fun c -> coreAddParent c state idx) children;
  if isUnsat then begin
    propagateUnsat [UState (state, Some idx)];
    false
  end else true

let rec insertAllRules state = function
  | [] -> true
  | (dep, chldrn)::tl ->
      let notUnsat = insertRule state dep chldrn in
      if notUnsat then insertAllRules state tl else false

let expandState state =
  match stateNextRule state with
  | MultipleElements rules ->
      let notUnsat = insertAllRules state rules in
      if notUnsat then stateSetStatus state Open
  | SingleElement (dep, chldrn) ->
      let notUnsat = insertRule state dep chldrn in
      if notUnsat then queueInsertState state else ()
  | NoMoreElements -> stateSetStatus state Open


let expandCnstr cset =
  let nht = nhtInit () in
  let mkCores f =
    let (sort, lf) as nom = atFormulaGetNominal f in
    let nomset =
      match nhtFind nht nom with
      | Some ns -> ns
      | None ->
          let cset1 = csetCopy cset in
          csetRemDot cset1;
          let bs = csetAddTBox sort cset1 in
          bsetAdd bs lf;
          nhtAdd nht nom bs;
          bs
    in
    bsetAdd nomset (atFormulaGetSubformula f)
  in
  csetIter cset mkCores;
  let inCores (sort, _) bs (isUns, acc) =
    let core = insertCore sort bs bs in (* TODO: think of deferral / μ stuff here *)
    coreAddConstraintParent core cset;
    (coreGetStatus core = Unsat || isUns, core::acc)
  in
  let (isUnsat, children) = nhtFold inCores nht (false, []) in
  if isUnsat then propagateUnsat [UCnstr cset]
  else
    match graphFindCnstr cset with
    | Some (UnexpandedC parents) -> graphReplaceCnstr cset (OpenC parents)
    | _ -> assert false


(*****************************************************************************)
(*                           The Main Loop                                   *)
(*****************************************************************************)

let expandNodesLoop (recursiveAction: unit -> unit) =
  match queueGetElement () with
  | QState state ->
      if stateGetStatus state = Expandable then begin
        expandState state;
        if doNominalPropagation () then begin
          (* propagateNominals (); *)
          if doSatPropagation () then (propagateUnsatMu (); propagateSatMu ())
        end else ()
      end else ();
      recursiveAction ()
  | QCore core ->
      if coreGetStatus core = Expandable then begin
        expandCore core;
        if doNominalPropagation () then begin
          (* propagateNominals (); *)
          if doSatPropagation () then (propagateUnsatMu (); propagateSatMu ())
        end else ()
      end else ();
      recursiveAction ()
  | QCnstr cset ->
      expandCnstr cset;
      recursiveAction ()
  | QEmpty -> ()

let runReasonerStep () =
  let void () = () in
  expandNodesLoop void; (* expand at most one node *)
  (* if this emptied the queue *)
  if queueIsEmpty () then begin
    (* then check whether the nominals would add further queue elements *)
    (* print_endline "Propagating nominals..."; *)
    (* propagateNominals () *)
  end else () (* else: the next step would be to expand another node *)

let rec buildGraphLoop () =
  let rec expandNodesFurther () = expandNodesLoop expandNodesFurther in
  expandNodesFurther (); (* expand as many queue elements as possible *)
  (* propagateNominals (); *)
  (* if propagating nominals added some more queue members, do all again.. *)
  if queueIsEmpty () then () else buildGraphLoop ()

let initReasoner sorts nomTable tbox sf =
  sortTable := Array.copy sorts;
  let (tbox1, sf1, bs) = ppFormulae nomTable tbox sf in
  let sort = fst sf in
  let bs1 = bsetAddTBox sort bs in
  let deferrals = bsetMakeRealEmpty() in
  let markDeferral f =
    if (Hashtbl.hash "ε") != (lfGetDeferral sort f) then (
      bsetAdd deferrals f;)
  in
  bsetIter markDeferral bs;
  graphInit ();
  queueInit ();
  let root = insertCore sort bs1 deferrals in
  graphAddRoot root

let isRootSat () =
  match coreGetStatus (graphGetRoot ()) with
  | Expandable -> None
  | Unsat -> Some false
  | Sat -> Some true
  | Open -> if (queueIsEmpty()) then Some true else None

let reasonerFinished () =
  match coreGetStatus (graphGetRoot ()) with
  | Expandable -> false
  | Unsat
  | Sat  -> true
  | Open -> queueIsEmpty ()


(** A graph-tableau-based decision procedure framework for coalgebraic logics.
    @param verbose An optional switch which determines
    whether the procedure shall print some information on the standard output.
    The default is false.
    @param sorts An array mapping each sort (represented as an integer)
    to a functor and a list of sorts which are the arguments of the functor.
    @param nomTable A partial function mapping nominals (represented as strings)
    to their sorts.
    @param tbox A list of sorted formulae.
    @param sf A sorted formula.
    @return True if sf is satisfiable wrt tbox, false otherwise.
 *)

let isSat ?(verbose = false) sorts nomTable tbox sf =
  let start = if verbose then Unix.gettimeofday () else 0. in
  initReasoner sorts nomTable tbox sf;
  let sat =
    try
      buildGraphLoop ();
      propagateUnsatMu ();
      (* get whether the root is satisfiable *)
      (* we know that the reasoner finished, so the value is there, *)
      (* i.e. isRootSat() will give a "Some x" and not "None" *)
      CU.fromSome (isRootSat())
    with CoAlg_finished res -> res
  in
  (* print some statistics *)
  if verbose then
    let stop = Unix.gettimeofday () in
    let addup lst = List.fold_left (fun acc sf -> acc + (CoAlgFormula.sizeSortedFormula sf)) 0 lst in
    print_newline ();
    print_endline ("Query: " ^ (CoAlgFormula.exportQuery tbox sf));
    let size = (CoAlgFormula.sizeSortedFormula sf) + (addup tbox) in
    print_endline ("Added Size: " ^ (string_of_int size));
    (*
    print_endline ("Negation Normal Form: " ^ (CoAlgFormula.exportQuery tbox1 sf1));
    let nsize = (CoAlgFormula.sizeSortedFormula sf1) + (addup tbox1) in
    print_endline ("Added Size: " ^ (string_of_int nsize));
    *)
    print_endline ("Result: Query is " ^ (if not sat then "not " else "") ^ "satisfiable.");
    print_endline ("Time: " ^ (string_of_float (stop -. start)));
    print_endline ("Generated states: " ^ (string_of_int (graphSizeState ())));
    print_endline ("Generated cores: " ^ (string_of_int (graphSizeCore ())));
    print_endline ("Generated constraints: " ^ (string_of_int (graphSizeCnstr ())));
    print_newline ()
  else ();
  sat
